
// Simple I/O

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include "adlib.h"


static char* findFormat(char* control) { 

  // Basically return pointer to first char after '%', but
  // skipping "%%" substrings.

  char* ptr = control - 1 ;
  do {
    ptr++ ;
    if(!(ptr = strchr(ptr, '%'))) return 0 ;
    ptr++ ;
  } while(*ptr == '%') ;
  return ptr ;
}

static void AdlibBroadcast(const int blockSize, char* data) {
  TreeSchedule tree ;
  spanTree(&tree, apg) ;
  tree.broadcast(blockSize, data) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// gprintf                                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void gprintf(char* _control, ...) {

  // Analogue of `printf' for globally replicated values

  if(AdlibLocId() == 0) {

    // Copy control string (copy eventually left unchanged, but will use it
    // as a scratch area, and `_control' might be in an unwritable segment).

    // (Later introduction of `control_' is a bug-fix that ought to be
    // be cleaned up.  Need to save this so we can delete it.)

    char* control_ = new char[strlen(_control) + 1] ;

    char* control = control_ ;
    strcpy(control, _control) ;

    // Process the arguments.

    va_list ap ;
    char* ptr ;
     
    va_start(ap, _control) ;

    while(ptr = findFormat(control)) {

      // Process a format specification.

      int scanning = 1 ;
      int nmods = 0, width, prcsn ;
      int isLong = 0, isShort = 0, isLongDouble = 0 ;
      do {
	switch(*ptr) {
	  case 0 : 

	    // let `printf' report an error, then go home.

	    printf(control) ;
	    delete [] control_ ;
	    return ;
	  case '*' :
	    if(nmods == 0) width = va_arg(ap, int) ;
	    if(nmods == 1) prcsn = va_arg(ap, int) ;
	    nmods++ ;
	    ptr++ ;
	    break ;
	  case 'l' :
	    isLong = 1 ;
	    ptr++ ;
	    break ;
	  case 'h' :
	    isShort = 1 ;
	    ptr++ ;
	    break ;
	  case 'L' :
	    isLongDouble = 1 ;
	    ptr++ ;
	    break ;
	  case 'c' : {
	    char value = va_arg(ap, char) ;
	    ptr++ ;
	    char save = *ptr ;
	    *ptr = 0 ;
	    if(nmods == 0) printf(control, value) ;
	    if(nmods == 1) printf(control, width, value) ;
	    if(nmods == 2) printf(control, width, prcsn, value) ;
	    *ptr = save ;
	    control = ptr ;
	    scanning = 0 ;
	    break ;
	  }
	  case 'd' :
	  case 'i' :
	  case 'o' :
	  case 'u' :
	  case 'x' :
	  case 'X' : {
	    ptr++ ;
	    char save = *ptr ;
	    *ptr = 0 ;
	    if(isLong) {
	      long value = va_arg(ap, long) ;
	      if(nmods == 0) printf(control, value) ;
	      if(nmods == 1) printf(control, width, value) ;
	      if(nmods == 2) printf(control, width, prcsn, value) ;
	    }
	    else if(isShort) {
	      short value = va_arg(ap, short) ;
	      if(nmods == 0) printf(control, value) ;
	      if(nmods == 1) printf(control, width, value) ;
	      if(nmods == 2) printf(control, width, prcsn, value) ;
	    }
	    else {
	      int value = va_arg(ap, int) ;
	      if(nmods == 0) printf(control, value) ;
	      if(nmods == 1) printf(control, width, value) ;
	      if(nmods == 2) printf(control, width, prcsn, value) ;
	    }
	    *ptr = save ;
	    control = ptr ;
	    scanning = 0 ;
	    break ;
	  }
	  case 'e' :
	  case 'E' :
	  case 'f' :
	  case 'g' :
	  case 'G' : {
	    double value = va_arg(ap, double) ;
	    ptr++ ;
	    char save = *ptr ;
	    *ptr = 0 ;
	    if(nmods == 0) printf(control, value) ;
	    if(nmods == 1) printf(control, width, value) ;
	    if(nmods == 2) printf(control, width, prcsn, value) ;
	    *ptr = save ;
	    control = ptr ;
	    scanning = 0 ;
	    break ;
	  }
	  case 's' :
	  case 'p' :
	  case 'n' : {
	    void* value = va_arg(ap, void*) ;
	    ptr++ ;
	    char save = *ptr ;
	    *ptr = 0 ;
	    if(nmods == 0) printf(control, value) ;
	    if(nmods == 1) printf(control, width, value) ;
	    if(nmods == 2) printf(control, width, prcsn, value) ;
	    *ptr = save ;
	    control = ptr ;
	    scanning = 0 ;
	    break ;
	  }
	  default :
	    ptr++ ;
	}
      } while(scanning) ;
    }

    va_end(ap) ;

    // Print trailing string

    printf(control) ;

    delete [] control_ ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// gscanf                                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void gscanf(char* _control, ...) {

  // Analogue of `scanf' for globally replicated values

  char* control_ = new char[strlen(_control) + 1] ;

  char* control = control_ ;
  strcpy(control, _control) ;

  // Process the arguments.

  va_list ap ;
  char* ptr ;
   
  va_start(ap, _control) ;

  while(ptr = findFormat(control)) {

    // Process a format specification.

    int scanning = 1 ;
    int ignoreInput = 0 ;
    int isLong = 0, isShort = 0, isLongDouble = 0 ;
    do {
      switch(*ptr) {
	case 0 : 

	  // let `scanf' report an error, then go home.

	  scanf(control) ;
	  delete [] control_ ;
	  return ;
	case '*' :
	  ignoreInput = 1 ;
	  ptr++ ;
	  break ;
	case 'l' :
	  isLong = 1 ;
	  ptr++ ;
	  break ;
	case 'h' :
	  isShort = 1 ;
	  ptr++ ;
	  break ;
	case 'L' :
	  isLongDouble = 1 ;
	  ptr++ ;
	  break ;
	case 'c' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
	  if(!ignoreInput) {
	    char* variable = va_arg(ap, char*) ;
	    if(AdlibLocId() == 0)
	      scanf(control, variable) ;
	    AdlibBroadcast(1, variable) ;
	  }
	  else
	    if(AdlibLocId() == 0)
	      scanf(control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 'd' :
	case 'i' :
	case 'o' :
	case 'u' :
	case 'x' :
	case 'X' :
	case 'n' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
	  if(!ignoreInput) {
	    if(isLong) {
	      long* variable = va_arg(ap, long*) ;
	      if(AdlibLocId() == 0)
		scanf(control, variable) ;
	      AdlibBroadcast(sizeof(long), (char*) variable) ;
	    }
	    else if(isShort) {
	      short* variable = va_arg(ap, short*) ;
	      if(AdlibLocId() == 0)
		scanf(control, variable) ;
	      AdlibBroadcast(sizeof(short), (char*) variable) ;
	    }
	    else {
	      int* variable = va_arg(ap, int*) ;
	      if(AdlibLocId() == 0)
		scanf(control, variable) ;
	      AdlibBroadcast(sizeof(int), (char*) variable) ;
	    }
	  }
	  else
	    if(AdlibLocId() == 0)
	      scanf(control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 'e' :
	case 'E' :
	case 'f' :
	case 'g' :
	case 'G' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
	  if(!ignoreInput) {
	    if(isLong) {
	      double* variable = va_arg(ap, double*) ;
	      if(AdlibLocId() == 0)
		scanf(control, variable) ;
	      AdlibBroadcast(sizeof(double), (char*) variable) ;
            }
            else if(isLongDouble) {
	      long double* variable = va_arg(ap, long double*) ;
	      if(AdlibLocId() == 0)
		scanf(control, variable) ;
	      AdlibBroadcast(sizeof(long double), (char*) variable) ;
            }
            else {
	      float* variable = va_arg(ap, float*) ;
	      if(AdlibLocId() == 0)
		scanf(control, variable) ;
	      AdlibBroadcast(sizeof(float), (char*) variable) ;
            }
          }
          else
            if(AdlibLocId() == 0)
	      scanf(control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	case 's' :
	case ']' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
          if(!ignoreInput) {
	    char* variable = va_arg(ap, char*) ;
            int len ;
            if(AdlibLocId() == 0) {
	      scanf(control, variable) ;
              len = strlen(variable) ;
            }
            AdlibBroadcast(sizeof(int), (char*) &len) ;
            AdlibBroadcast(len, variable) ;
          }
          else
            if(AdlibLocId() == 0)
	      scanf(control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	case 'p' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
          if(!ignoreInput) {
	    void** variable = va_arg(ap, void**) ;
            if(AdlibLocId() == 0)
	      scanf(control, variable) ;
            AdlibBroadcast(sizeof(void*), (char*) variable) ;
          }
          else
            if(AdlibLocId() == 0)
	      scanf(control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	default :
	  ptr++ ;
      }
    } while(scanning) ;
  }

  va_end(ap) ;

  // Skip trailing string

  if(AdlibLocId() == 0)
    scanf(control) ;

  delete [] control_ ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// aprintf                                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct aprintfArg {
  enum Type {CHAR, INT, FLOAT, PTR, NEWLN, RNG} ;

  aprintfArg(Type _type, const DAD* _value, void* _data, const int _len,
             char* _control, char* end,
	     int _nmods, int* _mods) ;

  aprintfArg(Type _type, int d,
             char* _control, char* end,
             int _nmods, int* _mods) ;

  aprintfArg(Type _type, int decim, 
             char* _control, char* end) ;

  ~aprintfArg() ;

  int type  ;

  const DAD* value ;
  void* data ;
  int len ;
  int d ;
  int decim ;

  char* control ;

  int nmods ;
  int mods [2] ;

  aprintfArg* next ;
} ;

aprintfArg :: aprintfArg(Type _type,
                         const DAD* _value, void* _data, const int _len,
                         char* _control, char* end,
			 int _nmods, int* _mods) {

  // `end' will point to char after the complete format specifier.

  type  = _type  ;
  value = _value ;
  data  = _data  ;
  len   = _len   ;

  int strlen = end - _control ;
  control = new char [strlen + 1] ;
  for(int i = 0 ; i < strlen ; i++)
    control [i] = _control [i] ;
  control [strlen] = 0 ;
  
  nmods = _nmods ;
  for(int j = 0 ; j < nmods ; j++)
    mods [j] = _mods [j] ;
}

aprintfArg :: aprintfArg(Type _type, int _d, char* _control, char* end,
			 int _nmods, int* _mods) {

  // _type == RNG

  // `end' will point to char after `R', ie the dimension selector.

  type  = _type ;
  value = 0 ;
  d     = _d ;

  int strlen = end - _control - 1 ;
  control = new char [strlen + 1] ;
  for(int i = 0 ; i < strlen - 1 ; i++)
    control [i] = _control [i] ;
  control [strlen - 1] = 'd' ;
  control [strlen]     = 0   ;

  nmods = _nmods ;
  for(int j = 0 ; j < nmods ; j++)
    mods [j] = _mods [j] ;
}

aprintfArg :: aprintfArg(Type, int _decim, char* _control, char* end) {

  // _type == NEWLN

  // `end' will point to char after `%', typically the start of the
  // ``field width''.

  type  = NEWLN ;
  value = 0 ;
  decim = _decim ;

  int strlen = end - _control + 1 ;
  control = new char [strlen + 1] ;
  for(int i = 0 ; i < strlen ; i++)
    control [i] = _control [i] ;
  control [strlen - 1] = 's' ;
  control [strlen] = 0 ;
  
  nmods = 0 ;
}

aprintfArg :: ~aprintfArg() {
  delete [] control ;
}

static void aprintfLoop(int narrs, char** arrsData, int rank, int* idxs,
                        int d, int* el,
                        aprintfArg* args, char* tail,
                        const DAD* arr, int* arrsLen) {

  if(d == rank) {
    aprintfArg* arg = args ;
    int j = 0 ;
    while(arg) {
      char* control = arg->control ;
      int nmods = arg->nmods ;
      int* mods = arg->mods ;
      int type  = arg->type ;
      switch(type) {
        case aprintfArg :: CHAR : {
	  char value = *((char*) (arrsData [j])) ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, mods [0], value) ;
	  if(nmods == 2) printf(control, mods [0], mods [1], value) ;
	  j++ ;
	  break ;
        }
        case aprintfArg :: INT : {
	  int value = *((int *) (arrsData [j])) ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, mods [0], value) ;
	  if(nmods == 2) printf(control, mods [0], mods [1], value) ;
	  j++ ;
	  break ;
        }
        case aprintfArg :: FLOAT : {
	  double value ;
          if(arrsLen [j] == sizeof(float))
	    value = *((float *) (arrsData [j])) ;
          else
	    value = *((double *) (arrsData [j])) ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, mods [0], value) ;
	  if(nmods == 2) printf(control, mods [0], mods [1], value) ;
	  j++ ;
	  break ;
        }
        case aprintfArg :: PTR : {
	  void* value = *((void**) (arrsData [j])) ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, mods [0], value) ;
	  if(nmods == 2) printf(control, mods [0], mods [1], value) ;
	  j++ ;
	  break ;
        }
        case aprintfArg :: RNG : {
	  int value = idxs [arg->d] ;
	  if(nmods == 0) printf(control, value) ;
	  if(nmods == 1) printf(control, mods [0], value) ;
	  if(nmods == 2) printf(control, mods [0], mods [1], value) ;
          break ;
        }
        case aprintfArg :: NEWLN : {
          int decim = arg->decim ;
	  if(decim == 0 || *el % decim == decim - 1)
	    printf(control, "\n") ;
	  else
	    printf(control, "") ;
	  break ;
        }
      }
      arg = arg->next ;
    }
    printf(tail) ;
    (*el)++ ;
  }
  else {
    Range x = arr->rng(d) ;
    const int count = x.size() ;

    int j ;

    char** arrsDataElem = new char* [narrs] ;
    int*   arrsStep     = new int   [narrs] ;
    for(j = 0 ; j < narrs ; j++) {
      arrsDataElem [j] = (char*) arrsData [j] ;
      arrsStep [j]     = arrsLen [j] * arr->str(d) ;
    }

    for(int sub = 0 ; sub < count ; sub++) {
      idxs [d] = sub ;

      aprintfLoop(narrs, arrsDataElem, rank, idxs, d + 1, el,
                  args, tail, arr, arrsLen) ;

      for(j = 0 ; j < narrs ; j++)
	arrsDataElem [j] += arrsStep [j] ;
    }

    delete [] arrsStep ;
    delete [] arrsDataElem ;
  }
}


void aprintf(char* control, IOArg* args) {

  // Analogue of `printf' for distributed array values

  int narrs = 0 ;
  aprintfArg* fields ;
  char* tail ;

  // Process the arguments.

  aprintfArg** argTail = &fields ;

  char* ptr ;
   
  while(ptr = findFormat(control)) {

    // Process a format specification.

    int scanning = 1 ;
    int nmods = 0 ;
    int mods [2] ;
    int decim = 0 ;
    char* formatStart = ptr ;
    do {
      switch(*ptr) {
	case 0 : 
          AdlibError("aprintf: unexpected end of format string") ;
	case '*' :
	  mods [nmods] = args->iValue ;
          args++ ;
	  nmods++ ;
	  ptr++ ;
	  break ;
	case 'c' : {
	  ptr++ ;
          
          aprintfArg* arg = new aprintfArg(aprintfArg :: CHAR,
                                     args->aValue, args->aData, args->aLen,
                                     control, ptr, nmods, mods) ;
          args++ ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
          narrs++ ; 

	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 'd' :
	case 'i' :
	case 'o' :
	case 'u' :
	case 'x' :
	case 'X' : {
	  ptr++ ;

          aprintfArg* arg = new aprintfArg(aprintfArg :: INT,
                                     args->aValue, args->aData, args->aLen,
                                     control, ptr, nmods, mods) ;
          args++ ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
          narrs++ ; 

	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 'e' :
	case 'E' :
	case 'f' :
	case 'g' :
	case 'G' : {
	  ptr++ ;

          aprintfArg* arg = new aprintfArg(aprintfArg :: FLOAT,
                                     args->aValue, args->aData, args->aLen,
                                     control, ptr, nmods, mods) ;
          args++ ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
          narrs++ ; 

	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 's' :
	case 'p' :
	case 'n' : {
	  ptr++ ;

          aprintfArg* arg = new aprintfArg(aprintfArg :: PTR,
                                     args->aValue, args->aData, args->aLen,
                                     control, ptr, nmods, mods) ;
          args++ ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
          narrs++ ; 

	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
        case 'R' : {
          ptr++ ;
          int d = *ptr - '0' ;

          ptr++ ;

          aprintfArg* arg = new aprintfArg(aprintfArg :: RNG, d,
                                           control, ptr, nmods, mods) ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
           
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
        case 'N' : {
          ptr++ ;

          if(nmods > 0) decim = mods [0] ;
          aprintfArg* arg = new aprintfArg(aprintfArg :: NEWLN, decim,
                                           control, formatStart) ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
           
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	default : {
          if('0' <= *ptr & *ptr <= '9')
	    decim = 10 * decim + (*ptr - '0') ;
		// Untidy hack.  Value of `decim' only used if format
                // eventually turns out to be `N'.
	  ptr++ ;
        }
      }
    } while(scanning) ;
  }

  *argTail = 0 ;

  tail = control ;

  // Remap argument arrays...

  Procs root(0, 0) ;
  int rank ; 
  DAD* dst ;
  int size ;
   
  char** arrsData = new char* [narrs] ;
  int*   arrsLen  = new int   [narrs] ;

  aprintfArg* arg = fields ;
  int j = 0 ;
  while(arg) {
    const DAD* src = arg->value ;
    if(src) {
      if(j == 0) {
	// Set `dst' to a collapsed frame, same shape as `src'.

        rank = src->rnk() ;
	dst  = new DAD(rank, root) ;
        dst->maps = new Map [rank] ;
        size = 1 ;
	for(int d = 0 ; d < rank ; d++) {
	  CollapsedRange x(src->rng(d).size()) ;
	  dst->maps [d] = Map(x, size) ;
          size *= x.volume() ;
	}
      }

      const int len = arg->len ;
      char* dstData = new char [len * size] ;

      // Remap `arg->data' to `dstData'.

      Remap schedule(dst, src, len) ;
      schedule.execute(dstData, arg->data) ;

      arrsData [j] = dstData ;
      arrsLen  [j] = len     ;

      j++ ;
    }
    arg = arg->next ;
  }

  // Do the printing...

  int* idxs = new int [rank] ;
  int el = 0 ;

  if(root.member()) {
    aprintfLoop(narrs, arrsData, rank, idxs, 0, &el, fields, tail,
                dst, arrsLen) ;
  }

  delete [] idxs ;

  for(j = narrs - 1 ; j >= 0 ; j--)
    delete [] arrsData [j] ;

  delete [] dst->maps ;
  delete dst ;

  delete [] arrsLen  ;
  delete [] arrsData ;

  // Clean up.

  arg = fields ;
  while(arg) {
    aprintfArg* next = arg->next ;
    delete arg ;
    arg = next ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// ascanf                                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct ascanfArg {

  ascanfArg(const DAD* _value, void* _data, const int _len,
            char* _control, char* end) ;
  ascanfArg(char* _control, char* end) ;

  ~ascanfArg() ;

  int ignore ;
  const DAD* value ;
  void* data ;
  int len ;

  char* control ;

  ascanfArg* next ;
} ;

ascanfArg :: ascanfArg(const DAD* _value, void* _data, const int _len,
                       char* _control, char* end) {

  // `end' will point to char after the complete format specifier.

  ignore = 0 ;
  value  = _value ;
  data   = _data ;
  len    = _len ;

  int strlen = end - _control ;
  control = new char [strlen + 1] ;
  for(int i = 0 ; i < strlen ; i++)
    control [i] = _control [i] ;
  control [strlen] = 0 ;
}

ascanfArg :: ascanfArg(char* _control, char* end) {

  // `end' will point to char after the complete format specifier.

  ignore = 1 ;

  int strlen = end - _control ;
  control = new char [strlen + 1] ;
  for(int i = 0 ; i < strlen ; i++)
    control [i] = _control [i] ;
  control [strlen] = 0 ;
}

ascanfArg :: ~ascanfArg() {
  delete [] control ;
}

static void ascanfLoop(int narrs, char** arrsData, int rank,
                       int d,
                       ascanfArg* args, char* tail,
                       const DAD* arr, int* arrsLen) {

  if(d == rank) {
    ascanfArg* arg = args ;
    int j = 0 ;
    while(arg) {
      int   ignore  = arg->ignore ;
      char* control = arg->control ;

      if(!ignore) {
	scanf(control, (void*) (arrsData [j])) ;
	j++ ;
      }
      else
	scanf(control) ;

      arg = arg->next ;
    }
    scanf(tail) ;
  }
  else {
    Range x = arr->rng(d) ;
    const int count = x.size() ;

    int j ;

    char** arrsDataElem = new char* [narrs] ;
    int*   arrsStep     = new int   [narrs] ;
    for(j = 0 ; j < narrs ; j++) {
      arrsDataElem [j] = arrsData [j] ;
      arrsStep [j]     = arrsLen [j] * arr->str(d) ;
    }

    for(int sub = 0 ; sub < count ; sub++) {
      ascanfLoop(narrs, arrsDataElem, rank, d + 1, args, tail, arr, arrsLen) ;

      for(j = 0 ; j < narrs ; j++)
	arrsDataElem [j] += arrsStep [j] ;
    }

    delete [] arrsStep ;
    delete [] arrsDataElem ;
  }
}

void ascanf(char* control, IOArg* args) {

  // Analogue of `scanf' for distributed array values

  int narrs = 0 ;
  ascanfArg* fields ;
  char* tail ;

  // Process the arguments.

  ascanfArg** argTail = &fields ;

  char* ptr ;
   
  while(ptr = findFormat(control)) {

    // Process a format specification.

    int scanning = 1 ;
    int ignoreInput = 0 ;
    do {
      switch(*ptr) {
	case 0 : 
          AdlibError("ascanf: unexpected end of format string") ;
	case '*' : {
          ignoreInput = 1 ;
	  ptr++ ;
	  break ;
        }
	case 'c' :
	case 'd' :
	case 'i' :
	case 'o' :
	case 'u' :
	case 'x' :
	case 'X' :
	case 'n' :
	case 'e' :
	case 'E' :
	case 'f' :
	case 'g' :
	case 'G' :
	case 's' :
	case ']' :
	case 'p' : {
	  ptr++ ;
          
	  ascanfArg* arg ;
          if(!ignoreInput) {
	    arg = new ascanfArg(args->aValue, args->aData, args->aLen,
                                control, ptr) ;
            args++ ;
            narrs++ ; 
          }
          else
	    arg = new ascanfArg(control, ptr) ;
          *argTail = arg ;
          argTail  = &(arg->next) ;

	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	default :
	  ptr++ ;
      }
    } while(scanning) ;
  }

  *argTail = 0 ;

  tail = control ;

  // Build source arrays.

  Procs root(0, 0) ;
  int rank ;
  DAD* src ;
  int size ;

  char** arrsData = new char* [narrs] ;
  int*   arrsLen  = new int   [narrs] ;

  ascanfArg* arg = fields ;
  int j = 0 ;
  while(arg) {
    const DAD* dst = arg->value ;
    if(dst) {
      if(j == 0) {    
 
        // Set `src' to a collapsed DAD, same shape as `dst'.

	rank = dst->rnk() ;
        src  = new DAD(rank, root) ;
	src->maps = new Map [rank] ;
	size = 1 ;
	for(int d = 0 ; d < rank ; d++) {
	  CollapsedRange x(dst->rng(d).size()) ;
	  src->maps [d] = Map(x, size) ;
          size *= x.volume() ;
	}
      }

      const int len = arg->len ;
      arrsData [j] = new char [len * size] ;
      arrsLen  [j] = len     ;

      j++ ;
    }
    arg = arg->next ;
  }

  // Do the scanning...

  if(root.member())
    ascanfLoop(narrs, arrsData, rank, 0, fields, tail, src, arrsLen) ;

  // Remap src arrays to arguments...

  arg = fields ;
  j = 0 ;
  while(arg) {
    const DAD* dst = arg->value ;
    if(dst) {
      Remap schedule(dst, src, arrsLen [j]) ;
      schedule.execute(arg->data, arrsData [j]) ;

      j++ ;
    }
    arg = arg->next ;
  }

  // Clean up.

  for(j = narrs - 1 ; j >= 0 ; j--) {
    //delete [] arrsData [j] ;
    char* ptr = arrsData [j] ;  // g++ v2.7.2 needs this
    delete [] ptr ;
  }
  delete [] arrsLen  ;
  delete [] arrsData ;

  delete [] src->maps ;
  delete src ;

  arg = fields ;
  while(arg) {
    ascanfArg* next = arg->next ;
    delete arg ;
    arg = next ;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// gfprintf                                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void gfprintf(FILE* fd, char* _control, ...) {

  // Analogue of `fprintf' for globally replicated values

  if(AdlibLocId() == 0) {

    // Copy control string (copy eventually left unchanged, but will use it
    // as a scratch area, and `_control' might be in an unwritable segment).

    // (Later introduction of `control_' is a bug-fix that ought to be
    // be cleaned up.  Need to save this so we can delete it.)

    char* control_ = new char[strlen(_control) + 1] ;

    char* control = control_ ;
    strcpy(control, _control) ;

    // Process the arguments.

    va_list ap ;
    char* ptr ;
     
    va_start(ap, _control) ;

    while(ptr = findFormat(control)) {

      // Process a format specification.

      int scanning = 1 ;
      int nmods = 0, width, prcsn ;
      do {
	switch(*ptr) {
	  case 0 : 
            AdlibError("gprintf: unexpected end of format string") ;
	  case '*' :
	    if(nmods == 0) width = va_arg(ap, int) ;
	    if(nmods == 1) prcsn = va_arg(ap, int) ;
	    nmods++ ;
	    ptr++ ;
	    break ;
	  case 'c' : {
	    char value = va_arg(ap, char) ;
	    ptr++ ;
	    char save = *ptr ;
	    *ptr = 0 ;
	    if(nmods == 0) fprintf(fd, control, value) ;
	    if(nmods == 1) fprintf(fd, control, width, value) ;
	    if(nmods == 2) fprintf(fd, control, width, prcsn, value) ;
	    *ptr = save ;
	    control = ptr ;
            scanning = 0 ;
            break ;
	  }
	  case 'd' :
	  case 'i' :
	  case 'o' :
	  case 'u' :
	  case 'x' :
	  case 'X' : {
	    int value = va_arg(ap, int) ;
	    ptr++ ;
	    char save = *ptr ;
	    *ptr = 0 ;
	    if(nmods == 0) fprintf(fd, control, value) ;
	    if(nmods == 1) fprintf(fd, control, width, value) ;
	    if(nmods == 2) fprintf(fd, control, width, prcsn, value) ;
	    *ptr = save ;
	    control = ptr ;
            scanning = 0 ;
            break ;
	  }
	  case 'e' :
	  case 'E' :
	  case 'f' :
	  case 'g' :
	  case 'G' : {
	    double value = va_arg(ap, double) ;
	    ptr++ ;
	    char save = *ptr ;
	    *ptr = 0 ;
	    if(nmods == 0) fprintf(fd, control, value) ;
	    if(nmods == 1) fprintf(fd, control, width, value) ;
	    if(nmods == 2) fprintf(fd, control, width, prcsn, value) ;
	    *ptr = save ;
	    control = ptr ;
            scanning = 0 ;
            break ;
	  }
	  case 's' :
	  case 'p' :
	  case 'n' : {
	    void* value = va_arg(ap, void*) ;
	    ptr++ ;
	    char save = *ptr ;
	    *ptr = 0 ;
	    if(nmods == 0) fprintf(fd, control, value) ;
	    if(nmods == 1) fprintf(fd, control, width, value) ;
	    if(nmods == 2) fprintf(fd, control, width, prcsn, value) ;
	    *ptr = save ;
	    control = ptr ;
            scanning = 0 ;
            break ;
	  }
	  default :
	    ptr++ ;
	}
      } while(scanning) ;
    }

    va_end(ap) ;

    // Print trailing string

    fprintf(fd, control) ;

    delete [] control_ ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// gfscanf                                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void gfscanf(FILE* fd, char* _control, ...) {

  // Analogue of `fscanf' for globally replicated values

  char* control_ = new char[strlen(_control) + 1] ;

  char* control = control_ ;
  strcpy(control, _control) ;

  // Process the arguments.

  va_list ap ;
  char* ptr ;
   
  va_start(ap, _control) ;

  while(ptr = findFormat(control)) {

    // Process a format specification.

    int scanning = 1 ;
    int ignoreInput = 0 ;
    do {
      switch(*ptr) {
	case 0 : 
          AdlibError("gscanf: unexpected end of format string") ;
	case '*' :
          ignoreInput = 1 ;
	  ptr++ ;
	  break ;
	case 'c' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
          if(!ignoreInput) {
	    char* variable = va_arg(ap, char*) ;
            if(AdlibLocId() == 0)
	      fscanf(fd, control, variable) ;
            AdlibBroadcast(1, variable) ;
          }
          else
            if(AdlibLocId() == 0)
	      fscanf(fd, control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	case 'd' :
	case 'i' :
	case 'o' :
	case 'u' :
	case 'x' :
	case 'X' :
	case 'n' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
          if(!ignoreInput) {
	    int* variable = va_arg(ap, int*) ;
            if(AdlibLocId() == 0)
	      fscanf(fd, control, variable) ;
            AdlibBroadcast(sizeof(int), (char*) variable) ;
          }
          else
            if(AdlibLocId() == 0)
	      fscanf(fd, control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	case 'e' :
	case 'E' :
	case 'f' :
	case 'g' :
	case 'G' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
          if(!ignoreInput) {
	    double* variable = va_arg(ap, double*) ;
            if(AdlibLocId() == 0)
	      fscanf(fd, control, variable) ;
            AdlibBroadcast(sizeof(double), (char*) variable) ;
          }
          else
            if(AdlibLocId() == 0)
	      fscanf(fd, control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	case 's' :
	case ']' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
          if(!ignoreInput) {
	    char* variable = va_arg(ap, char*) ;
            int len ;
            if(AdlibLocId() == 0) {
	      fscanf(fd, control, variable) ;
              len = strlen(variable) ;
            }
            AdlibBroadcast(sizeof(int), (char*) &len) ;
            AdlibBroadcast(len, variable) ;
          }
          else
            if(AdlibLocId() == 0)
	      fscanf(fd, control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	case 'p' : {
	  ptr++ ;
	  char save = *ptr ;
	  *ptr = 0 ;
          if(!ignoreInput) {
	    void** variable = va_arg(ap, void**) ;
            if(AdlibLocId() == 0)
	      fscanf(fd, control, variable) ;
            AdlibBroadcast(sizeof(void*), (char*) variable) ;
          }
          else
            if(AdlibLocId() == 0)
	      fscanf(fd, control) ;
	  *ptr = save ;
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	default :
	  ptr++ ;
      }
    } while(scanning) ;
  }

  va_end(ap) ;

  // Skip trailing string

  if(AdlibLocId() == 0)
    fscanf(fd, control) ;

  delete [] control_ ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// afprintf                                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

static void afprintfLoop(int narrs, char** arrsData, int rank, int* idxs,
                         int d, int* el,
                         aprintfArg* args, char* tail,
                         const DAD* arr, int* arrsLen, FILE* fd) {

  if(d == rank) {
    aprintfArg* arg = args ;
    int j = 0 ;
    while(arg) {
      char* control = arg->control ;
      int nmods = arg->nmods ;
      int* mods = arg->mods ;
      int type  = arg->type ;
      switch(type) {
        case aprintfArg :: CHAR : {
	  char value = *((char*) (arrsData [j])) ;
	  if(nmods == 0) fprintf(fd, control, value) ;
	  if(nmods == 1) fprintf(fd, control, mods [0], value) ;
	  if(nmods == 2) fprintf(fd, control, mods [0], mods [1], value) ;
	  j++ ;
	  break ;
        }
        case aprintfArg :: INT : {
	  int value = *((int *) (arrsData [j])) ;
	  if(nmods == 0) fprintf(fd, control, value) ;
	  if(nmods == 1) fprintf(fd, control, mods [0], value) ;
	  if(nmods == 2) fprintf(fd, control, mods [0], mods [1], value) ;
	  j++ ;
	  break ;
        }
        case aprintfArg :: FLOAT : {
	  double value ;
          if(arrsLen [j] == sizeof(float))
	    value = *((float *) (arrsData [j])) ;
          else
	    value = *((double *) (arrsData [j])) ;
	  if(nmods == 0) fprintf(fd, control, value) ;
	  if(nmods == 1) fprintf(fd, control, mods [0], value) ;
	  if(nmods == 2) fprintf(fd, control, mods [0], mods [1], value) ;
	  j++ ;
	  break ;
        }
        case aprintfArg :: PTR : {
	  void* value = *((void**) (arrsData [j])) ;
	  if(nmods == 0) fprintf(fd, control, value) ;
	  if(nmods == 1) fprintf(fd, control, mods [0], value) ;
	  if(nmods == 2) fprintf(fd, control, mods [0], mods [1], value) ;
	  j++ ;
	  break ;
        }
        case aprintfArg :: RNG : {
	  int value = idxs [arg->d] ;
	  if(nmods == 0) fprintf(fd, control, value) ;
	  if(nmods == 1) fprintf(fd, control, mods [0], value) ;
	  if(nmods == 2) fprintf(fd, control, mods [0], mods [1], value) ;
          break ;
        }
        case aprintfArg :: NEWLN : {
          int decim = arg->decim ;
	  if(decim == 0 || *el % decim == decim - 1)
	    fprintf(fd, control, "\n") ;
	  else
	    fprintf(fd, control, "") ;
	  break ;
        }
      }
      arg = arg->next ;
    }
    fprintf(fd, tail) ;
    (*el)++ ;
  }
  else {
    Range x = arr->rng(d) ;
    const int count = x.size() ;

    int j ;

    char** arrsDataElem = new char* [narrs] ;
    int*   arrsStep     = new int   [narrs] ;
    for(j = 0 ; j < narrs ; j++) {
      arrsDataElem [j] = (char*) arrsData [j] ;
      arrsStep [j]     = arrsLen [j] * arr->str(d) ;
    }

    for(int sub = 0 ; sub < count ; sub++) {
      idxs [d] = sub ;

      afprintfLoop(narrs, arrsDataElem, rank, idxs, d + 1, el,
                   args, tail, arr, arrsLen, fd) ;

      for(j = 0 ; j < narrs ; j++)
	arrsDataElem [j] += arrsStep [j] ;
    }

    delete [] arrsStep ;
    delete [] arrsDataElem ;
  }
}


void afprintf(FILE* fd, char* control, IOArg* args) {

  // Analogue of `printf' for distributed array values

  int narrs = 0 ;
  aprintfArg* fields ;
  char* tail ;

  // Process the arguments.

  aprintfArg** argTail = &fields ;

  char* ptr ;
   
  while(ptr = findFormat(control)) {

    // Process a format specification.

    int scanning = 1 ;
    int nmods = 0 ;
    int mods [2] ;
    int decim = 0 ;
    char* formatStart = ptr ;
    do {
      switch(*ptr) {
	case 0 : 
          AdlibError("aprintf: unexpected end of format string") ;
	case '*' :
	  mods [nmods] = args->iValue ;
          args++ ;
	  nmods++ ;
	  ptr++ ;
	  break ;
	case 'c' : {
	  ptr++ ;
          
          aprintfArg* arg = new aprintfArg(aprintfArg :: CHAR,
                                     args->aValue, args->aData, args->aLen,
                                     control, ptr, nmods, mods) ;
          args++ ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
          narrs++ ; 

	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 'd' :
	case 'i' :
	case 'o' :
	case 'u' :
	case 'x' :
	case 'X' : {
	  ptr++ ;

          aprintfArg* arg = new aprintfArg(aprintfArg :: INT,
                                     args->aValue, args->aData, args->aLen,
                                     control, ptr, nmods, mods) ;
          args++ ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
          narrs++ ; 

	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 'e' :
	case 'E' :
	case 'f' :
	case 'g' :
	case 'G' : {
	  ptr++ ;

          aprintfArg* arg = new aprintfArg(aprintfArg :: FLOAT,
                                     args->aValue, args->aData, args->aLen,
                                     control, ptr, nmods, mods) ;
          args++ ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
          narrs++ ; 

	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
	case 's' :
	case 'p' :
	case 'n' : {
	  ptr++ ;

          aprintfArg* arg = new aprintfArg(aprintfArg :: PTR,
                                     args->aValue, args->aData, args->aLen,
                                     control, ptr, nmods, mods) ;
          args++ ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
          narrs++ ; 

	  control = ptr ;
	  scanning = 0 ;
	  break ;
	}
        case 'R' : {
          ptr++ ;
          int d = *ptr - '0' ;

          ptr++ ;

          aprintfArg* arg = new aprintfArg(aprintfArg :: RNG, d,
                                           control, ptr, nmods, mods) ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
           
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
        case 'N' : {
          ptr++ ;

          if(nmods > 0) decim = mods [0] ;
          aprintfArg* arg = new aprintfArg(aprintfArg :: NEWLN, decim,
                                           control, formatStart) ;
          *argTail = arg ;
          argTail  = &(arg->next) ;
           
	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	default : {
          if('0' <= *ptr & *ptr <= '9')
	    decim = 10 * decim + (*ptr - '0') ;
		// Untidy hack.  Value of `decim' only used if format
                // eventually turns out to be `N'.
	  ptr++ ;
        }
      }
    } while(scanning) ;
  }

  *argTail = 0 ;

  tail = control ;

  // Remap argument arrays...

  Procs root(0, 0) ;
  int rank ; 
  DAD* dst ;
  int size ;
   
  char** arrsData = new char* [narrs] ;
  int*   arrsLen  = new int   [narrs] ;

  aprintfArg* arg = fields ;
  int j = 0 ;
  while(arg) {
    const DAD* src = arg->value ;
    if(src) {
      if(j == 0) {
	// Set `dst' to a collapsed frame, same shape as `src'.

        rank = src->rnk() ;
	dst  = new DAD(rank, root) ;
	dst->maps = new Map [rank] ;
	size = 1 ;
	for(int d = 0 ; d < rank ; d++) {
	  CollapsedRange x(src->rng(d).size()) ;
	  dst->maps [d] = Map(x, size) ;
          size *= x.volume() ;
	}
      }

      const int len = arg->len ;
      char* dstData = new char [len * size] ;

      // Remap `arg->data' to `dstData'.

      Remap schedule(dst, src, len) ;
      schedule.execute(dstData, arg->data) ;
      
      arrsData [j] = dstData ;
      arrsLen  [j] = len     ;

      j++ ;
    }
    arg = arg->next ;
  }

  // Do the printing...

  int* idxs = new int [rank] ;
  int el = 0 ;

  if(root.member())
    afprintfLoop(narrs, arrsData, rank, idxs, 0, &el, fields, tail,
                 dst, arrsLen, fd) ;

  delete [] idxs ;

  for(j = narrs - 1 ; j >= 0 ; j--)
    delete [] arrsData [j] ;

  delete [] dst->maps ;
  delete dst ;

  delete [] arrsLen  ;
  delete [] arrsData ;

  // Clean up.

  arg = fields ;
  while(arg) {
    aprintfArg* next = arg->next ;
    delete arg ;
    arg = next ;
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// afscanf                                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

static void afscanfLoop(int narrs, char** arrsData, int rank,
                        int d,
                        ascanfArg* args, char* tail,
                        const DAD* arr, int* arrsLen, FILE* fd) {

  if(d == rank) {
    ascanfArg* arg = args ;
    int j = 0 ;
    while(arg) {
      int   ignore  = arg->ignore ;
      char* control = arg->control ;

      if(!ignore) {
	fscanf(fd, control, (void*) (arrsData [j])) ;
	j++ ;
      }
      else
	fscanf(fd, control) ;

      arg = arg->next ;
    }
    scanf(tail) ;
  }
  else {
    Range x = arr->rng(d) ;
    const int count = x.size() ;

    int j ;

    char** arrsDataElem = new char* [narrs] ;
    int*   arrsStep     = new int   [narrs] ;
    for(j = 0 ; j < narrs ; j++) {
      arrsDataElem [j] = arrsData [j] ;
      arrsStep [j]     = arrsLen [j] * arr->str(d) ;
    }

    for(int sub = 0 ; sub < count ; sub++) {
      afscanfLoop(narrs, arrsDataElem, rank, d + 1, args, tail,
                  arr, arrsLen, fd) ;

      for(j = 0 ; j < narrs ; j++)
	arrsDataElem [j] += arrsStep [j] ;
    }

    delete [] arrsStep ;
    delete [] arrsDataElem ;
  }
}

void afscanf(FILE* fd, char* control, IOArg* args) {

  // Analogue of `scanf' for distributed array values

  int narrs = 0 ;
  ascanfArg* fields ;
  char* tail ;

  // Process the arguments.

  ascanfArg** argTail = &fields ;

  char* ptr ;
   
  while(ptr = findFormat(control)) {

    // Process a format specification.

    int scanning = 1 ;
    int ignoreInput = 0 ;
    do {
      switch(*ptr) {
	case 0 : 
          AdlibError("afscanf: unexpected end of format string") ;
	case '*' : {
          ignoreInput = 1 ;
	  ptr++ ;
	  break ;
        }
	case 'c' :
	case 'd' :
	case 'i' :
	case 'o' :
	case 'u' :
	case 'x' :
	case 'X' :
	case 'n' :
	case 'e' :
	case 'E' :
	case 'f' :
	case 'g' :
	case 'G' :
	case 's' :
	case ']' :
	case 'p' : {
	  ptr++ ;
          
	  ascanfArg* arg ;
          if(!ignoreInput) {
	    arg = new ascanfArg(args->aValue, args->aData, args->aLen,
                                control, ptr) ;
            args++ ;
            narrs++ ; 
          }
          else
	    arg = new ascanfArg(control, ptr) ;
          *argTail = arg ;
          argTail  = &(arg->next) ;

	  control = ptr ;
	  scanning = 0 ;
	  break ;
        }
	default :
	  ptr++ ;
      }
    } while(scanning) ;
  }

  *argTail = 0 ;

  tail = control ;

  // Build source arrays.

  Procs root(0, 0) ;
  int rank ;
  DAD* src ;
  int size ;

  char** arrsData = new char* [narrs] ;
  int*   arrsLen  = new int   [narrs] ;

  ascanfArg* arg = fields ;
  int j = 0 ;
  while(arg) {
    const DAD* dst = arg->value ;
    if(dst) {
      if(j == 0) {    
 
        // Set `src' to a collapsed DAD, same shape as `dst'.

	rank = dst->rnk() ;
        src  = new DAD(rank, root) ;
	src->maps = new Map [rank] ;
	size = 1 ;
	for(int d = 0 ; d < rank ; d++) {
	  CollapsedRange x(dst->rng(d).size()) ;
	  src->maps [d] = Map(x, size) ;
          size *= x.volume() ;
	}
      }

      const int len = arg->len ;
      arrsData [j] = new char [len * size] ;
      arrsLen  [j] = len     ;

      j++ ;
    }
    arg = arg->next ;
  }

  // Do the scanning...

  if(root.member())
    afscanfLoop(narrs, arrsData, rank, 0, fields, tail, src, arrsLen, fd) ;

  // Remap src arrays to arguments...

  arg = fields ;
  j = 0 ;
  while(arg) {
    const DAD* dst = arg->value ;
    if(dst) {
      Remap schedule(dst, src, arrsLen [j]) ;
      schedule.execute(arg->data, arrsData [j]) ;

      j++ ;
    }
    arg = arg->next ;
  }

  // Clean up.

  for(j = narrs - 1 ; j >= 0 ; j--) {
    //delete [] arrsData [j] ;
    char* ptr = arrsData [j] ;  // g++ v2.7.2 needs this
    delete [] ptr ;
  }
  delete [] arrsLen  ;
  delete [] arrsData ;

  delete [] src->maps ;
  delete src ;

  arg = fields ;
  while(arg) {
    ascanfArg* next = arg->next ;
    delete arg ;
    arg = next ;
  }
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// gfopen                                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

FILE* gfopen(char* name, char* mode) {
  FILE *result = 0 ;

  Procs root(0, 0) ;
  if(root.member())
    result = fopen(name, mode) ;

  return result ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// gfclose                                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

void gfclose(FILE* fd) {

  Procs root(0, 0) ;
  if(root.member())
    fclose(fd) ;
}


// -------------------------------------------------------------
// Things to do
//
//   Change `aprintf', `ascanf', ... to deal with `h', `l', `L' modifiers
//   (already done for `gprintf', `gscanf')
//
//   Add `lprintf', which is functionally equivalent to existing `AdlibPrintf'.
//
//   `aprintf' etc should check for zero data fields in the control string.
//
//   Is it worth generalizing format options to support things like
//   complex arrays.  Perhaps you want some sort of nested format like
//
//     Array2<complex> a ;
//     ...
//     aprintf("%[(%f %f)%]\n", a)
//
//   Won't work because don't know the byte-lengths of individual
//   %f fields in element.  Would first have to solve problem of
//
//     complex a ;
//     ...
//     printf("%[(%f %f)%]\n", a)
//
//   Better add some field in format descriptor to describe byte offset
//   in argument...
//     
//     aprintf("%f %[4]f\n", a, a) ;
//
//   which would also allow
//
//     aprintf("%f %[*]f\n", a, sizeof(float), a) ;
//
//   `AdlibBroadcast' should be renamed.

